from typing import Iterable

import numpy as np
from pixaris.metrics.base import BaseMetric
from PIL.Image import Image


def _luminescence(image: Image) -> np.array:
    """
    calculates the luminescence values for an image
    definition luminance: https://www.101computing.net/colour-luminance-and-contrast-ratio/

    :param image: input Image
    :type image: Image
    :return: luminescence values per pixel, normed between 0 and 1
    :rtype: np.array
    """
    image = np.asarray(image) / 255
    image = np.where(image <= 0.03928, image / 12.92, ((image + 0.055) / 1.055) ** 2.4)
    return 0.2126 * image[:, :, 0] + 0.7152 * image[:, :, 1] + 0.0722 * image[:, :, 2]


class LuminescenceComparisonByMaskMetric(BaseMetric):
    """
    Calculate the luminescence difference between the masked part and the unmasked part of an image.
    The metric is calculated as the absolute difference between the average luminescence of the masked part
    and the unmasked part of the image.
    The result is a number between 0 and 1. 1 is the best possible score (minimal difference in luminescence of masked and unmasked part) and 0 is the worst score (maximal difference).
    """

    def __init__(
        self,
        mask_images: Iterable[Image],
    ):
        super().__init__()
        self.mask_images = mask_images

    def _luminescence_difference(self, image: Image, mask: Image) -> float:
        """
        Calculate the differences of luminescence in the masked part and the unmasked part of the image.
        a number close to 1 means the luminescence is close, a number close to 0 means the luminescence is very different.

        :param image: The input image.
        :type image: Image.Image
        :param mask: The mask image.
        :type mask: Image.Image
        :return: The luminescence difference value of the image.
        :rtype: float
        """
        binary_mask = np.array(mask.convert("L").point(lambda p: p > 125 and 255)) / 255
        inverted_mask = 1 - binary_mask

        luminescence = _luminescence(image)
        mean_masked_luminescence = np.average(luminescence, weights=binary_mask)
        mean_inverted_luminescence = np.average(luminescence, weights=inverted_mask)
        return 1 - abs(
            mean_masked_luminescence - mean_inverted_luminescence
        )  # natural a number between 0 and 1

    def calculate(self, generated_images: Iterable[Image]) -> dict:
        """
        Calculate the luminescence score of a list of generated images.
        For each image we calculate the average luminescence of the masked part and the unmasked part,
        and return the absolute difference between them. Luminescence is a number between 0 and 1, so
        the result is also a number between 0 and 1. We invert them to make 1 the best score (minimal difference in luminescence of masked and unmasked part) and 0 the worst (maximal difference).

        :param generated_images: A list of generated images.
        :type generated_images: Iterable[Image]
        :return: A dictionary containing a single entry: "luminescence_difference": the average luminescence_difference score.
        :rtype: dict
        """
        luminescence_scores = []
        for gen, mask in zip(generated_images, self.mask_images):
            brightness_difference = self._luminescence_difference(gen, mask)
            luminescence_scores.append(brightness_difference)

        return {
            "luminescence_difference": np.mean(luminescence_scores)
            if luminescence_scores
            else 0
        }


class LuminescenceWithoutMaskMetric(BaseMetric):
    """
    Calculates mean and variance of the luminescence of the image.
    """

    def _luminescence_mean_and_var(self, image: Image) -> float:
        """
        Calculate the mean and variance of the luminescence of the image.

        :param image: The input image.
        :type image: Image.Image
        :return: mean and variance of the luminescence
        :rtype: tuple[float, float]
        """
        luminescence = _luminescence(image)
        mean = np.mean(luminescence)
        var = np.var(luminescence)
        return [mean, var]

    def calculate(self, generated_images: Iterable[Image]) -> dict:
        """
        Calculate the luminescence score of a list of generated images.
        For each image we calculate the mean and variance of the luminescence,
        and return the average of them. The results is 2 numbers between 0 and 1

        :param generated_images: A list of generated images.
        :type generated_images: Iterable[Image]
        :return: A dictionary containing different luminescence statistics:
        :rtype: dict
        """
        luminescence_scores = []
        for gen in generated_images:
            brightness_difference = self._luminescence_mean_and_var(gen)
            luminescence_scores.append(brightness_difference)

        mean_values = np.mean(np.array(luminescence_scores), axis=0)
        return {"luminescence_mean": mean_values[0], "luminescence_var": mean_values[1]}



import gradio as gr
import os
from pixaris.frontend.tab_feedback import render_feedback_tab
from pixaris.frontend.tab_experiment_tracking import render_experiment_tracking_tab
from pixaris.feedback_handlers.base import FeedbackHandler
from pixaris.experiment_handlers.base import ExperimentHandler


def launch_ui(
    feedback_handler: FeedbackHandler,
    experiment_handler: ExperimentHandler,
    results_directory="local_results/",
    server_name="localhost",
):
    """
    Launches the Gradio UI for Pixaris.
    Args:
        feedback_handler: The feedback handler to use.
        experiment_handler: The experiment handler to use.
        results_directory: The directory to save the results in.
        server_name: The name of the server to launch the UI on. Set "localhost" for local testing and "0.0.0.0" for app engine deployment.
    """
    with gr.Blocks(
        title="Pixaris",
        theme=gr.themes.Monochrome(
            spacing_size=gr.themes.sizes.spacing_sm, radius_size="none"
        ),
        analytics_enabled=False,
        fill_width=False,
    ) as demo:
        # add title
        with gr.Row():
            with gr.Column(scale=1, min_width=100):
                gr.Image(
                    value="assets/pixaris_logo.png",
                    elem_id="pixaris-logo",
                    show_label=False,
                    height=100,
                    width=100,
                    show_download_button=False,
                    show_fullscreen_button=False,
                    show_share_button=False,
                    interactive=False,
                )
            with gr.Column(scale=20):
                gr.Markdown(
                    """
                    <div style="display: flex; align-items: center; height: 100%; justify-content: flex-start;">
                        <h1 style="margin-left: 20px;">Pixaris</h1>
                    </div>
                    """
                )
        with gr.Tab("Experiment Tracking"):
            render_experiment_tracking_tab(
                experiment_handler=experiment_handler,
                results_directory=results_directory,
            )

        with gr.Tab("Feedback"):
            render_feedback_tab(
                feedback_handler=feedback_handler,
            )
    demo.launch(
        server_name=server_name, server_port=8080, allowed_paths=[os.path.abspath("./")]
    )



import json
import shutil
from pixaris.feedback_handlers.base import FeedbackHandler
import gradio as gr
from datetime import datetime
import os
import pandas as pd


class LocalFeedbackHandler(FeedbackHandler):
    """
    Local feedback handler for Pixaris. This class is used to handle feedback
    iterations locally. It allows for the creation of feedback iterations,
    writing feedback to local storage, and loading feedback data from local
    storage.

    :param project_feedback_dir: Directory where feedback iterations are stored. Default is "feedback_iterations".
    :type project_feedback_dir: str
    :param project_feedback_file: Name of the feedback file. Default is "feedback_tracking.jsonl".
    :type project_feedback_file: str
    :param local_feedback_directory: Local directory for storing feedback data. Default is "local_results".
    :type local_feedback_directory: str
    """

    def __init__(
        self,
        project_feedback_dir: str = "feedback_iterations",
        project_feedback_file: str = "feedback_tracking.jsonl",
        local_feedback_directory: str = "local_results",
    ):
        os.makedirs(local_feedback_directory, exist_ok=True)
        self.local_feedback_directory = local_feedback_directory
        self.project_feedback_dir = project_feedback_dir
        self.project_feedback_file = project_feedback_file
        self.feedback_df = None
        self.feedback_iteration_choices = None
        self.projects = None

    def write_single_feedback(self, feedback: dict) -> None:
        """
        Writes feedback for one image to local feedback storage.

        :param feedback: dict with feedback information. Dict is expected to have the following keys:
        * project: name of the project
        * feedback_iteration: name of the iteration
        * dataset: name of the evaluation set (optional)
        * image_name: name of the image
        * experiment_name: name of the experiment (optional)
        * feedback_indicator: string with feedback value (either "Like", "Dislike", or "Neither")
        * comment: string with feedback comment (optional)
        :type feedback: dict
        """
        row_to_insert = self._construct_feedback_row_to_insert(feedback)

        # Construct the directory and file path
        project_dir = os.path.join(
            self.local_feedback_directory,
            feedback["project"],
        )
        os.makedirs(project_dir, exist_ok=True)
        feedback_file_path = os.path.join(
            project_dir, "feedback_iterations", self.project_feedback_file
        )

        # Write feedback to the file
        with open(feedback_file_path, "a") as feedback_file:
            feedback_file.write(json.dumps(row_to_insert) + "\n")

        # Display success message
        gr.Info("Feedback saved locally", duration=1)

    def _save_images_to_feedback_iteration_folder(
        self,
        local_image_directory: str,
        project: str,
        feedback_iteration: str,
    ):
        """
        Copies images from the experiment directory to the feedback directory.

        :param local_image_directory: Path to the directory containing the images for the feedback iteration
        :type local_image_directory: str
        :param project: Name of the project
        :type project: str
        :param feedback_iteration: Name of the feedback iteration
        :type feedback_iteration: str
        """
        # copy the image folder to the feedback directory
        feedback_dir = os.path.join(
            self.local_feedback_directory,
            project,
            self.project_feedback_dir,
            feedback_iteration,
        )
        os.makedirs(feedback_dir, exist_ok=True)
        shutil.copytree(
            local_image_directory,
            feedback_dir,
            dirs_exist_ok=True,
        )

    def _initialise_feedback_iteration_in_table(
        self,
        project: str,
        feedback_iteration: str,
        image_names: list[str],
        dataset: str = None,
        experiment_name: str = None,
    ):
        """
        Initialise feedback iteration locally in tracking table.

        :param project: Name of the project
        :type project: str
        :param feedback_iteration: Name of the feedback iteration
        :type feedback_iteration: str
        :param image_names: List of image names to use in the feedback iteration
        :type image_names: list[str]
        :param dataset: Name of the dataset (optional)
        :type dataset: str
        :param experiment_name: Name of the experiment (optional)
        :type experiment_name: str
        """

        # for each image, create the upload entry in feedback table
        for image_name in image_names:
            feedback = {
                "project": project,
                "feedback_iteration": feedback_iteration,
                "dataset": dataset,
                "experiment_name": experiment_name,
                "image_name": image_name,
                "feedback_indicator": "Neither",  # used only for initialisation of feedback iteration
                "comment": "upload",
            }
            self.write_single_feedback(feedback)

    def create_feedback_iteration(
        self,
        local_image_directory: str,
        project: str,
        feedback_iteration: str,
        date_suffix: str = None,
        dataset: str = "",
        experiment_name: str = "",
    ):
        """
        Saves images in experiment_directorey to a feedback_iteration folder.
        Puts initial entries into local feedback database.

        :param local_image_directory: Path to the directory containing the images
        :type local_image_directory: str
        :param project: Name of the project
        :type project: str
        :param feedback_iteration: Name of the feedback iteration
        :type feedback_iteration: str
        :param image_names: List of image names to use in the feedback iteration
        :type image_names: list[str]
        :param dataset: Name of the dataset (optional)
        :type dataset: str
        :param experiment_name: Name of the experiment (optional)
        :type experiment_name: str
        """
        # add date for versioning if not provided
        if not date_suffix:
            date_suffix = datetime.now().strftime("%y%m%d")
        feedback_iteration = f"{date_suffix}_{feedback_iteration}"

        self._save_images_to_feedback_iteration_folder(
            local_image_directory=local_image_directory,
            project=project,
            feedback_iteration=feedback_iteration,
        )

        image_names = os.listdir(local_image_directory)
        self._initialise_feedback_iteration_in_table(
            project=project,
            feedback_iteration=feedback_iteration,
            image_names=image_names,
            dataset=dataset,
            experiment_name=experiment_name,
        )

    def load_projects_list(self) -> list[str]:
        """
        Retrieves list of projects from local storage.

        :return: List of project names
        :rtype: list[str]
        """
        projects = [
            d
            for d in os.listdir(self.local_feedback_directory)
            if os.path.isdir(os.path.join(self.local_feedback_directory, d))
        ]
        projects.sort()
        self.projects = projects

        print(f"Found projects: {projects}")
        return projects

    def _load_feedback_df(self, project: str) -> pd.DataFrame:
        """
        Retrieves feedback data for a project from local storage. Adds paths for location of images in
        local directory to the dataframe.
        :param project: Name of the project
        :type project: str
        :return: DataFrame containing feedback data
        :rtype: pd.DataFrame
        """
        print(f"Searching locally for feedback data for project {project}...")
        feedback_file_path = os.path.join(
            self.local_feedback_directory,
            project,
            "feedback_iterations",
            self.project_feedback_file,
        )

        if not os.path.exists(feedback_file_path):
            raise FileNotFoundError(f"No feedback file found at {feedback_file_path}")

        # Load the feedback data from the local JSONL file
        feedback_data = []
        with open(feedback_file_path, "r") as feedback_file:
            for line in feedback_file:
                # Parse JSONL line into a dictionary
                feedback_data.append(json.loads(line.strip()))

        # Convert the feedback data to a DataFrame
        feedback_df = pd.DataFrame(feedback_data)

        # add local paths for images to feedback_df
        feedback_df["image_path_local"] = feedback_df.apply(
            lambda row: os.path.join(
                self.local_feedback_directory,
                row["project"],
                self.project_feedback_dir,
                row["feedback_iteration"],
                row["image_name"],
            ),
            axis=1,
        )

        # sum up likes and dislikes, split comments into liked and disliked
        feedback_df = (
            feedback_df.groupby(["project", "feedback_iteration", "image_name"])
            .agg(
                likes=("likes", lambda x: (x == 1).sum()),
                dislikes=("dislikes", lambda x: (x == 1).sum()),
                comments_liked=(
                    "comment",
                    lambda x: list(
                        x[(feedback_df["likes"] == 1) & (~x.isin(["", " "]))]
                    ),
                ),
                comments_disliked=(
                    "comment",
                    lambda x: list(
                        x[(feedback_df["dislikes"] == 1) & (~x.isin(["", " "]))]
                    ),
                ),
                image_path_local=("image_path_local", "first"),
            )
            .reset_index()
        )
        return feedback_df

    def load_images_for_feedback_iteration(
        self,
        feedback_iteration: str,
        sorting: str = "image_name",
    ) -> list[str]:
        """
        Returns list of local image paths that belong to the feedback iteration.

        :param feedback_iteration: Name of the feedback iteration
        :type feedback_iteration: str
        :param sorting: Sorting option for the images. Can be "image_name", "likes", or "dislikes". Default is "image_name".
        :type sorting: str
        :return: List of local image paths
        :rtype: list[str]
        """
        print(f"Loading images for feedback iteration {feedback_iteration}...")

        # get relevant data for this feedback iteration
        iteration_df = self.feedback_df.loc[
            # only this feedback iteration
            self.feedback_df["feedback_iteration"] == feedback_iteration
        ].copy()

        # sort images according to user preference
        sorted_df = self._sort_iteration_df(
            iteration_df=iteration_df,
            sorting=sorting,
        )
        # deduplicate image paths
        image_paths_local = sorted_df["image_path_local"].unique().tolist()
        return image_paths_local



import gradio as gr
from pixaris.feedback_handlers.base import FeedbackHandler

PROJECTS = []


def render_feedback_tab(
    feedback_handler: FeedbackHandler,
):
    # initially load all projects
    global PROJECTS
    PROJECTS = [""] + feedback_handler.load_projects_list()

    feedback_details = (
        gr.State(  # is adjusted from inside a gr.render decorated function. See below.
            value={
                "project": "",
                "feedback_iteration": "",
                "image_name": "",
                "feedback_indicator": False,
                "comment": "",
            }
        )
    )

    def adjust_feedback_details(
        img_path: str, feedback_indicator: bool, comment: str, previous_details: dict
    ):
        """
        This function is only here to adjust the feedback details (that are a gr.State object). This is
        necessary because the feedback details are determined within a function that has the gr.render
        decorator. Thus, the only way to trigger a function within the gr.render function (in this case
        we want to write the feedback) is to write the inputs to a gr.State object and then call the
        respective function within here.
        See here https://www.gradio.app/guides/state-in-blocks (see example with cart).
        """
        previous_details["project"] = img_path.split("/")[-4]  # todo make more robust
        previous_details["feedback_iteration"] = img_path.split("/")[-2]
        previous_details["image_name"] = img_path.split("/")[-1]
        previous_details["feedback_indicator"] = feedback_indicator
        previous_details["comment"] = comment
        feedback_handler.write_single_feedback(feedback=previous_details)
        return previous_details

    with gr.Row():
        with gr.Column(scale=2, min_width=250):
            project_name = gr.Dropdown(
                choices=PROJECTS,
                label="Project",
                filterable=True,
            )

            # initialise hidden feedback iterations and define updating function
            feedback_iterations = gr.Dropdown(visible=False)

            def update_feedback_iteration_choices(project_name, feedback_iterations):
                """Update choices of feedback iterations for selected project."""
                feedback_handler.load_all_feedback_iteration_info_for_project(
                    project_name
                )
                feedback_iteration_choices = feedback_handler.feedback_iteration_choices

                feedback_iterations = gr.Dropdown(
                    label="Feedback Iterations",
                    value=None,
                    choices=feedback_iteration_choices,
                    visible=True,
                    filterable=True,
                    multiselect=True,
                    max_choices=100,
                    interactive=True,
                )
                return feedback_iterations

            # update feedback iterations choices upon project change
            project_name.change(
                fn=update_feedback_iteration_choices,
                inputs=[
                    project_name,
                    feedback_iterations,
                ],
                outputs=[feedback_iterations],
            )

            reload_projects_button = gr.Button(
                "Reload projects",
                variant="secondary",
                interactive=True,
                size="sm",
            )

            def reload_projects(project_name):
                global PROJECTS
                PROJECTS = feedback_handler.load_projects_list()
                project_name = gr.Dropdown(
                    choices=PROJECTS,
                    label="Project",
                    filterable=True,
                )
                return project_name

            reload_projects_button.click(
                fn=reload_projects,
                inputs=[project_name],
                outputs=[project_name],
            )

            columns = gr.Slider(
                minimum=1,
                maximum=7,
                value=5,
                label="Number of images per row",
                step=1,
            )

            display_feedback_checkbox = gr.Checkbox(
                value=False,
                label="Show previous feedback",
                interactive=True,
            )
            sorting_of_images = gr.Radio(
                choices=["image_name", "likes", "dislikes"],
                label="Sorting of images",
                value="image_name",
            )
            reload_feedback_button = gr.Button(
                "Reload feedback",
                variant="secondary",
                interactive=True,
                size="sm",
            )
            reloaded_feedback_count = gr.State(value=1)

            def reload_feedback(project_name, reloaded_feedback_count):
                feedback_handler.load_all_feedback_iteration_info_for_project(
                    project_name
                )
                reloaded_feedback_count += 1
                gr.Info("Reloaded feedback.", duration=2)
                return reloaded_feedback_count

            reload_feedback_button.click(
                fn=reload_feedback,
                inputs=[project_name, reloaded_feedback_count],
                outputs=[reloaded_feedback_count],
            )

        # Main content on the right
        with gr.Column(scale=8):

            @gr.render(
                inputs=[
                    feedback_iterations,
                    columns,
                    display_feedback_checkbox,
                    sorting_of_images,
                    reloaded_feedback_count,
                ]
            )
            def render_images_per_iteration(
                feedback_iterations,
                columns,
                display_feedback_checkbox,
                sorting_of_images,
                reloaded_feedback_count,
            ):
                """
                This function renders the images for each feedback iteration. It is decorated with gr.render
                to allow for dynamic rendering of the images based on the selected feedback iterations.
                - for each feedback_iteration, there is a separate accordion
                - each accordion contains rows of images with the number of columns specified by the user
                - each image is associated with feedback functionality
                """
                if not feedback_iterations:
                    gr.Markdown("No feedback iteration selected.")
                    return
                for feedback_iteration in feedback_iterations:
                    # load the images corresponding to this feedback iteration
                    feedback_iteration_images = (
                        feedback_handler.load_images_for_feedback_iteration(
                            feedback_iteration, sorting_of_images
                        )
                    )

                    # split images into batches of number of columns
                    num_images = len(feedback_iteration_images)
                    images_batches = [
                        feedback_iteration_images[i : i + columns]
                        for i in range(0, num_images, columns)
                    ]

                    # fill up last batch with Nones to have same number of images in each row
                    if len(images_batches[-1]) < columns:
                        images_batches[-1] += [None] * (
                            columns - len(images_batches[-1])
                        )

                    # render images
                    min_width_elements = "10px"
                    with gr.Accordion(label=f"Iteration {feedback_iteration}"):
                        for batch in images_batches:
                            with gr.Row(variant="compact"):
                                for img in batch:
                                    # string of image_name is needed later on, img will be modified by gradio hereafter.
                                    img_path = str(img)
                                    img_name = img_path.split("/")[-1]
                                    # only display image with buttons if it exists
                                    element_visible_bool = bool(img)
                                    with gr.Column(
                                        variant="compact", min_width=min_width_elements
                                    ):
                                        gr.Image(
                                            value=img,
                                            label=img_name,
                                            show_download_button=True,
                                            show_fullscreen_button=True,
                                            visible=element_visible_bool,
                                            min_width=min_width_elements,
                                            scale=1,
                                        )
                                        img_textbox = gr.Textbox(
                                            value=img_path, visible=False
                                        )  # needed bc gr.render
                                        comment = gr.Textbox(
                                            label="Comment",
                                            value="",
                                            visible=element_visible_bool,
                                            min_width=min_width_elements,
                                            scale=1,
                                            interactive=True,
                                        )
                                        feedback_indicator = gr.Radio(
                                            choices=["Like", "Dislike"],
                                            label="Rating",
                                            visible=element_visible_bool,
                                        )
                                        feedback_button = gr.Button(
                                            "Save this image feedback",
                                            visible=element_visible_bool,
                                            size="sm",
                                            min_width=min_width_elements,
                                            scale=1,
                                            variant="primary",
                                            interactive=False,
                                        )

                                        # feedback button is only clickable if feedback indicator is changed recently
                                        def change_feedback_button_interactivity(
                                            feedback_indicator,
                                            feedback_button,
                                        ):
                                            if feedback_indicator:
                                                feedback_button = gr.Button(
                                                    interactive=True
                                                )
                                            else:
                                                feedback_button = gr.Button(
                                                    interactive=False
                                                )

                                            return feedback_button

                                        feedback_indicator.change(
                                            fn=change_feedback_button_interactivity,
                                            inputs=[
                                                feedback_indicator,
                                                feedback_button,
                                            ],
                                            outputs=[feedback_button],
                                        )

                                        # adjusts gr.State object with feedback details and writes feedback
                                        feedback_button.click(
                                            fn=adjust_feedback_details,
                                            inputs=[
                                                img_textbox,
                                                feedback_indicator,
                                                comment,
                                                feedback_details,
                                            ],
                                            outputs=[feedback_details],
                                        ).then(
                                            # change feedback button to non-interactive again (avoid double feedback)
                                            fn=lambda: gr.update(interactive=False),
                                            inputs=None,
                                            outputs=feedback_button,
                                        )

                                        feedback = (
                                            feedback_handler.get_feedback_per_image(
                                                feedback_iteration=feedback_iteration,
                                                image_name=img_name,
                                            )
                                        )
                                        gr.Markdown(
                                            label="Previous Feedback",
                                            value=f"Likes: {feedback['likes']} - Comments: {feedback['comments_liked']}"
                                            if feedback and reloaded_feedback_count > 0
                                            else "",
                                            visible=display_feedback_checkbox
                                            and element_visible_bool,
                                        )
                                        gr.Markdown(
                                            label="Previous Feedback",
                                            value=f"Dislikes: {feedback['dislikes']} - Comments: {feedback['comments_disliked']}"
                                            if feedback and reloaded_feedback_count > 0
                                            else "",
                                            visible=display_feedback_checkbox
                                            and element_visible_bool,
                                        )



from typing import Iterable
from PIL import Image


def dict_mean(input_dict_list: Iterable[dict]) -> dict:
    """
    Calculate the mean value for each key in a list of dictionaries.

    :param input_dict_list: A list of dictionaries with the same keys.
    :type input_dict_list: Iterable[dict]
    :return: A dictionary with the mean values for each key.
    :rtype: dict
    :raises ValueError: If the input list is empty or if the dictionaries have different keys.
    """
    # check if all dicts have the same keys
    if len(set(map(frozenset, input_dict_list))) > 1:
        raise ValueError("All dictionaries must have the same keys.")

    means_dict = {}
    for key in input_dict_list[0].keys():
        values = [d[key] for d in input_dict_list]
        means_dict[key] = sum(values) / len(values)
    return means_dict


def normalize_image(image: Image, max_size=(1024, 1024)) -> Image:
    """
    Normalize the given image by placing it on a white background, scaling it while preserving aspect ratio,
    and returning the resulting image.

    :param image: The input image to be normalized.
    :type image: PIL.Image.Image
    :param max_size: The maximum size of the output image, defaults to (1024, 1024).
    :type max_size: tuple[int, int], optional
    :return: The normalized image with a white background.
    :rtype: PIL.Image.Image
    """
    # place on white background
    if image.mode != "RGBA":
        image = image.convert("RGBA")

    # Scale image while preserving aspect ratio
    image.thumbnail(max_size, resample=Image.LANCZOS)

    # Create white background
    white_base = Image.new("RGB", max_size, (255, 255, 255))

    # Paste object image onto white background
    offset = ((max_size[0] - image.size[0]) // 2, (max_size[1] - image.size[1]) // 2)
    white_base.paste(image, offset, image)
    return white_base



from google.cloud import bigquery


def python_type_to_bq_type(python_type):
    """
    Maps a Python data type to a corresponding BigQuery data type.

    :param python_type: The Python data type to map.
    :type python_type: type
    :return: The corresponding BigQuery data type as a string.
    :rtype: str
    """
    type_mapping = {
        str: "STRING",
        int: "INTEGER",
        float: "FLOAT",
        bool: "BOOLEAN",
        bytes: "BYTES",
    }
    return type_mapping.get(
        python_type, "STRING"
    )  # Default to STRING if type is unknown, such as datetime


def create_schema_from_dict(data_dict) -> list[bigquery.SchemaField]:
    """
    Creates a BigQuery schema from a dictionary of data.

    :param data_dict: A dictionary where keys are field names and values are field values.
    :type data_dict: dict
    :return: A list of BigQuery SchemaField objects.
    :rtype: list[bigquery.SchemaField]
    """
    schema = []
    for key, value in data_dict.items():
        field_type = python_type_to_bq_type(type(value))
        schema.append(
            bigquery.SchemaField(name=key, field_type=field_type, mode="NULLABLE")
        )
    return schema


def ensure_table_exists(
    table_ref: str,
    bigquery_input: dict,
    bigquery_client: bigquery.Client,
):
    """
    Ensures that the BigQuery table exists with the correct schema.
    :param table_ref: The reference to the BigQuery table.
    :type table_ref: str
    :param bigquery_input: The dictionary used to generate the schema.
    :type bigquery_input: dict
    :param bigquery_client: The BigQuery client.
    :type bigquery_client: bigquery.Client
    """
    schema = create_schema_from_dict(bigquery_input)

    try:
        table = bigquery_client.get_table(table_ref)
    except Exception as e:
        if "Not found: Table" in str(e):
            table = bigquery.Table(table_ref, schema=schema)
            bigquery_client.create_table(table)
            table = bigquery_client.get_table(table_ref)
            print(f"Created table {table_ref}.")
        else:
            raise e
